
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __GA_ENUMS_ENUM_TYPES_H__
#define __GA_ENUMS_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "../../avahi-0.8/avahi-gobject/ga-enums.h" */
GType ga_protocol_get_type (void);
#define GA_TYPE_PROTOCOL (ga_protocol_get_type())
GType ga_lookup_result_flags_get_type (void);
#define GA_TYPE_LOOKUP_RESULT_FLAGS (ga_lookup_result_flags_get_type())
GType ga_lookup_flags_get_type (void);
#define GA_TYPE_LOOKUP_FLAGS (ga_lookup_flags_get_type())
GType ga_resolver_event_get_type (void);
#define GA_TYPE_RESOLVER_EVENT (ga_resolver_event_get_type())
GType ga_browser_event_get_type (void);
#define GA_TYPE_BROWSER_EVENT (ga_browser_event_get_type())
G_END_DECLS

#endif /* __GA_ENUMS_ENUM_TYPES_H__ */

/* Generated data ends here */

