
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <ga-entry-group.h>
#include<ga-entry-group-enumtypes.h>

/* enumerations from "../../avahi-0.8/avahi-gobject/ga-entry-group.h" */
GType
ga_entry_group_state_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { GA_ENTRY_GROUP_STATE_UNCOMMITED, "GA_ENTRY_GROUP_STATE_UNCOMMITED", "GA_ENTRY_GROUP_STATE_UNCOMMITED" },
      { GA_ENTRY_GROUP_STATE_REGISTERING, "GA_ENTRY_GROUP_STATE_REGISTERING", "GA_ENTRY_GROUP_STATE_REGISTERING" },
      { GA_ENTRY_GROUP_STATE_ESTABLISHED, "GA_ENTRY_GROUP_STATE_ESTABLISHED", "GA_ENTRY_GROUP_STATE_ESTABLISHED" },
      { GA_ENTRY_GROUP_STATE_COLLISTION, "GA_ENTRY_GROUP_STATE_COLLISTION", "GA_ENTRY_GROUP_STATE_COLLISTION" },
      { GA_ENTRY_GROUP_STATE_FAILURE, "GA_ENTRY_GROUP_STATE_FAILURE", "GA_ENTRY_GROUP_STATE_FAILURE" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("GaEntryGroupState", values);
  }
  return etype;
}

/* Generated data ends here */

